"""
S3ben entry module for cli commands
"""

import multiprocessing
import multiprocessing.managers
import os
import pwd
import signal
from argparse import Namespace
from logging import getLogger
from multiprocessing.synchronize import Event as EventClass

from typing_extensions import TypeAlias

from s3ben.arguments import base_args
from s3ben.backup import BackupManager
from s3ben.config import parse_config
from s3ben.decorators import argument, command
from s3ben.logger import init_logger
from s3ben.rabbit import MqConnection, MqExQue, RabbitMQ
from s3ben.remapper import ResolveRemmaping
from s3ben.s3 import S3Config, S3Events
from s3ben.sentry import init_sentry

_logger = getLogger(__name__)
args = base_args()
subparser = args.add_subparsers(dest="subcommand")
Queue: TypeAlias = multiprocessing.Queue
Event: TypeAlias = EventClass


def main() -> None:
    """
    Entry point
    :raises ValueError: if config file not found
    :return: None
    """
    parsed_args = args.parse_args()
    if parsed_args.subcommand is None:
        args.print_help()
        return
    init_logger(name="s3ben", level=parsed_args.log_level)
    if os.path.isfile(parsed_args.sentry_conf):
        _logger.debug("Initializing sentry")
        init_sentry(config=parsed_args.sentry_conf)
    config = parse_config(parsed_args.config)
    parsed_args.func(config, parsed_args)


def run_remmaper(
    data_queue: Queue,
    end_event: Event,
    backup_root: str,
) -> None:
    """
    Function to start remmaper process

    :param multiprocessing.Queue data_queue: MP queue for data exchange
    :param multiprocessing.Event end_event: MP Event for finishing process
    :param str backup_root: Backup main directory
    :return: None
    """
    remap_resolver = ResolveRemmaping(backup_root=backup_root)
    remap_resolver.run(queue=data_queue, event=end_event)


def run_consumer(end_event: Event, data_queue: Queue, config: dict) -> None:
    """
    Function to start consumers
    """
    main_section = config.pop("s3ben")
    mq_section: dict = config.pop("amqp")
    mq_ex_queue = MqExQue(
        exchange=mq_section.pop("exchange"), queue=mq_section.pop("queue")
    )
    mq_connection = MqConnection(**mq_section)

    backup_root = main_section.pop("backup_root")
    s3 = config.pop("s3")
    s3.pop("exclude")
    s3_config = S3Config(**s3)
    s3_events = S3Events(
        config=s3_config,
        backup_root=backup_root,
    )
    backup = BackupManager(
        backup_root=backup_root,
        user=main_section.pop("user"),
        mq_conn=mq_connection,
        mq_queue=mq_ex_queue.queue,
    )
    backup.start_consumer(
        s3_client=s3_events, mp_data_queue=data_queue, mp_end_event=end_event
    )


def exclude_list(config: dict) -> list:
    """
    Function to return exclude list from config

    :param dict config: Configuration dictionary from parseconfig
    :rtype: list
    :return: List of excluded buckets
    """
    if "exclude" not in config["s3"].keys():
        return []
    exclude = config["s3"].pop("exclude").replace('"', "").replace("'", "").split(",")
    return [b.strip() for b in exclude]


def signal_handler(signal_no, stack_frame) -> None:
    """
    Function to hanle signlas
    :raises SystemExit: In all cases
    """
    raise SystemExit


@command(parent=subparser)  # type: ignore
def setup(config: dict, *_) -> None:
    """
    Cli command to add required cofiguration to s3 buckets and mq
    :param dict config: Parsed configuration dictionary
    :return: None
    """
    _logger.info("Checking backup root")
    main_section: dict = config.pop("s3ben")
    backup_root: str = main_section.pop("backup_root")
    user = pwd.getpwnam(main_section.pop("user"))
    if not os.path.exists(backup_root):
        os.mkdir(path=backup_root, mode=0o700)
        os.chown(path=backup_root, uid=user.pw_uid, gid=user.pw_gid)
    _logger.info("Setting up RabitMQ")
    mq_section = config.pop("amqp")
    mq_ex_queue = MqExQue(
        exchange=mq_section.pop("exchange"), queue=mq_section.pop("queue")
    )
    mq_connection = MqConnection(**mq_section)
    mq = RabbitMQ(conn_params=mq_connection)
    mq.prepare(
        exchange=mq_ex_queue.exchange,
        queue=mq_ex_queue.queue,
        routing_key=mq_ex_queue.exchange,
    )
    _logger.info("Setting up S3")
    exclude_buckets = exclude_list(config=config)
    s3 = config.pop("s3")
    s3_config = S3Config(**s3)
    s3_events = S3Events(config=s3_config)
    all_buckets = s3_events.get_admin_buckets()
    filtered_buckets = list(set(all_buckets) - set(exclude_buckets))
    s3_events.create_topic(
        config=mq_connection,
        exchange=mq_ex_queue.exchange,
    )
    for bucket in filtered_buckets:
        _logger.debug("Setting up bucket: %s", bucket)
        s3_events.create_notification(bucket=bucket, exchange=mq_ex_queue.exchange)


@command(
    [
        argument("--bucket", required=True, help="Bucket name which to sync", type=str),
        argument(
            "--transfers",
            help="Number of transfer processes, default: %(default)d",
            type=int,
            default=4,
        ),
        argument(
            "--checkers",
            help="Number of checker processes, default: %(default)d",
            type=int,
            default=4,
        ),
        argument("--skip-checksum", help="Skip checksum check", action="store_true"),
        argument("--skip-filesize", help="Skip filesize check", action="store_true"),
        argument(
            "--page-size",
            help="Bucket object page size, default: %(default)s",
            type=int,
            default=1000,
        ),
        # TODO: Not working currently, as progress data was changed, needs update
        argument(
            "--ui",
            help="Use experimental ui, default: %(default)s",
            action="store_true",
        ),
        argument(
            "--avg-interval",
            help="Amount of seconds for calculating avg speed, default: %(default)d",
            type=int,
            default=60,
        ),
        argument(
            "--update-interval",
            help="Progress bar update interval, default: %(default)d",
            type=int,
            default=1,
        ),
    ],
    parent=subparser,  # type: ignore
)
def sync(config: dict, parsed_args: Namespace):
    """
    Entry point for sync cli option
    """
    if not 2 <= parsed_args.avg_interval <= 60:
        _logger.error("Avg interval must be between 2 and 60")
        return
    _logger.debug("Initializing sync")
    s3 = config.pop("s3")
    s3.pop("exclude")
    s3_config = S3Config(**s3)
    backup_root = config["s3ben"].pop("backup_root")
    s3_events = S3Events(
        config=s3_config,
        backup_root=backup_root,
    )
    backup = BackupManager(
        backup_root=backup_root,
        user=config["s3ben"].pop("user"),
        s3_client=s3_events,
        curses=parsed_args.ui,
    )
    backup.sync_bucket(
        parsed_args.bucket,
        parsed_args.transfers,
        parsed_args.page_size,
        parsed_args.checkers,
        parsed_args.skip_checksum,
        parsed_args.skip_filesize,
        avg_interval=parsed_args.avg_interval,
        update_interval=parsed_args.update_interval,
    )


@command(
    [
        argument("--show-excluded", help="Show excluded buckets", action="store_true"),
        argument("--show-obsolete", help="Show obsolete bucklets", action="store_true"),
        argument(
            "--only-enabled",
            help="Show only backup enabled buckets",
            action="store_true",
        ),
        argument(
            "--sort",
            help="Sort results by select collump, default: %(default)s",
            choices=["bucket", "owner", "size", "objects"],
            default="bucket",
        ),
        argument(
            "--sort-reverse", help="Reverse order for sorting", action="store_true"
        ),
    ],
    parent=subparser,  # type: ignore
)
def buckets(config: dict, parsed_args: Namespace) -> None:
    """
    Cli command: buckets
    """
    _logger.debug("Listing buckets")
    exclude = exclude_list(config=config)
    s3 = config.pop("s3")
    s3_config = S3Config(**s3)
    backup_root = config["s3ben"].pop("backup_root")
    s3_events = S3Events(
        config=s3_config,
        backup_root=backup_root,
    )
    backup = BackupManager(
        backup_root=backup_root,
        user=config["s3ben"].pop("user"),
        s3_client=s3_events,
    )
    backup.list_buckets(
        exclude=exclude,
        show_excludes=parsed_args.show_excluded,
        show_obsolete=parsed_args.show_obsolete,
        only_enabled=parsed_args.only_enabled,
        sort=parsed_args.sort,
        sort_revers=parsed_args.sort_reverse,
    )


@command(
    [
        argument(
            "--days-keep",
            help="How long to keep, default: %(default)d",
            default=30,
            type=int,
        ),
        argument(
            "--force",
            action="store_true",
            help="Force removing all deleted objects",
        ),
    ],
    parent=subparser,  # type: ignore
)
def cleanup(config: dict, parsed_args: Namespace) -> None:
    """
    Cli function to call deleted items cleanup method
    from BackupManager
    """
    _logger.debug("Starting deleted items cleanup")
    backup_root = config["s3ben"].pop("backup_root")
    backup = BackupManager(
        backup_root=backup_root,
        user=config["s3ben"].pop("user"),
    )
    if parsed_args.days_keep == 0:
        if not parsed_args.force:
            _logger.error(
                "This will remove all moved objects, use --force if you want to do this anyway"
            )
            return
        _logger.warning("Removing ALL deleted items")
    backup.cleanup_deleted_items(days=parsed_args.days_keep + 1)


@command(
    [
        argument(
            "--consumers",
            type=int,
            default=4,
            help="Number of consumer processes (max limited to cpu cores), default: %(default)s",
        )
    ],
    parent=subparser,  # type: ignore
)
def consume(config: dict, parsed_args: Namespace) -> None:
    """
    Function to start/restart consumers and other needed processes

    :param str backup_root: Backup root
    :param int n_proc: number of consumer processes to start,
        default 4 or max numbers of cpu cores
    :return: None
    """
    signal.signal(signal.SIGTERM, signal_handler)
    signal.signal(signal.SIGINT, signal_handler)

    max_proc = multiprocessing.cpu_count()
    n_consumers = min(max_proc, parsed_args.consumers)
    backup_root = config["s3ben"].get("backup_root")
    processes = []
    with multiprocessing.managers.SyncManager() as p_manager:
        data_queue = p_manager.Queue()
        end_event = p_manager.Event()
        try:
            remapper_proc = multiprocessing.Process(
                target=run_remmaper,
                args=(
                    data_queue,
                    end_event,
                    backup_root,
                ),
                name="remmaper",
            )
            remapper_proc.start()
            processes.append(remapper_proc)
            for _ in range(n_consumers):
                consumer = multiprocessing.Process(
                    target=run_consumer,
                    args=(
                        end_event,
                        data_queue,
                        config,
                    ),
                )
                consumer.start()
                processes.append(consumer)
            for proc in processes:
                proc.join()
        except (KeyboardInterrupt, SystemExit):
            for proc in processes:
                proc.terminate()



"""
Module to collect and visualise progress
"""

import multiprocessing
import os
import time
from dataclasses import dataclass, field
from logging import getLogger
from multiprocessing.synchronize import Event as EventClass
from typing import Tuple

from typing_extensions import TypeAlias

from s3ben.helpers import convert_to_human

_logger = getLogger(__name__)
Queue: TypeAlias = multiprocessing.Queue
Event: TypeAlias = EventClass


@dataclass
class ProgressMarkers:
    """
    Dataclass for progress markers
    """

    current: list = field(default_factory=lambda: ["-", "\\", "|", "/"])
    done: str = field(default="█")
    filler: str = field(default=".")

    def current_marker(self) -> str:
        """
        Metod to get and rotate current marker
        """
        current_marker = self.current.pop(0)
        self.current.append(current_marker)
        return current_marker


class ProgressSpeed:
    """
    Class to calculate avg and estimate time
    """

    def __init__(self, avg_interval: int) -> None:
        self.__speed_history = [0.0 for _ in range(avg_interval)]
        self._speed: float = 0

    @property
    def speed(self) -> float:
        """
        Property to return current speed
        :rtype: int
        """
        return self._speed

    @speed.setter
    def speed(self, value) -> None:
        """
        Method to set current speed
        """
        self._speed += value

    @property
    def avg_speed(self) -> float:
        """
        Property to calculate avg speed
        :rtype: float
        """
        return round(sum(self.__speed_history) / len(self.__speed_history), 2)

    def update_speed_history(self) -> None:
        """
        Update speed history
        """
        self.__speed_history.pop(0)
        self.__speed_history.append(self._speed)
        self._speed = 0


class ProgressTimings:
    """
    Dataclass to calculate timings for progress
    """

    def __init__(self) -> None:
        self._start: float = time.perf_counter()
        self._current: float = time.perf_counter()

    @property
    def start(self) -> int:
        """
        Return start time
        :rtype: int
        """
        return int(self._start)

    @property
    def running(self) -> tuple:
        """
        Method to calculate and return running time from start
        :rtype: tuple
        :return: (hours, minuts, seconds)
        """
        self._current = time.perf_counter()
        running = self.current - self.start
        hours = int(running) // 3600
        minutes = int(running) // 60 % 60
        seconds = int(running) % 60
        return (hours, minutes, seconds)

    @property
    def current(self) -> int:
        """
        Property to return current running time
        :rtype: int
        """
        return int(self._current)


class ProgressV2:
    """
    Dataclass to represent progress in console
    """

    def __init__(self, total: int, avg_interval: int) -> None:
        self.times = ProgressTimings()
        self.markers = ProgressMarkers()
        self.speed = ProgressSpeed(avg_interval=avg_interval)
        self._total = total
        self._current: int = 0
        self._download: int = 0

    @property
    def total(self) -> int:
        """
        Property method for returning total value
        :rtype: int
        """
        return self._total

    @property
    def current(self) -> int:
        """
        Property method to return current value
        :rtype: int
        """
        return self._current

    @property
    def download(self) -> int:
        """
        Property method to return download value
        """
        return self._download

    @property
    def estimate(self) -> Tuple[int, int, int]:
        """
        Calculate estimate time left
        """
        try:
            left = (self.total - self.current) / self.speed.avg_speed
        except ZeroDivisionError:
            left = 0
        h = int(left) // 3600
        m = int(left) // 60 % 60
        s = int(left) % 60
        return h, m, s

    @property
    def percents_done(self) -> float:
        """
        Property to calculate and return percents done
        """
        return round(self.current * 100 / self.total, 2)

    def update_counters(self, value: dict) -> None:
        """
        Method to update current progress value
        """
        dl = 0
        vrf = 0
        if "dl" in value.keys():
            dl = value.pop("dl")
            self._download += dl
        else:
            vrf = value.pop("vrf")
        self._current += vrf + dl
        self.speed.speed = dl + vrf


class ConsoleBar:
    """
    Class used to decorate console
    """

    def __init__(self, total: int, avg_interval: int, exit_event_queue: Event) -> None:
        self.data = ProgressV2(total=total, avg_interval=avg_interval)
        t, u = convert_to_human(self.data.total)
        self.__terminal_columns: int = os.get_terminal_size().columns
        self.__total = f"T {t:.2f}{u}" if u else f"T {t}"
        self.__progress: str = f"[V {0:7}|D {0:7}|{self.__total}]"
        self.__time: str = f"[R {0:02}:{0:02}:{0:02}][E {99:02}:{59:02}:{59:02}]"
        self.__percents: str = f"[{0:6.2f}%]"
        self.__avg_speed: str = f"[{0:>6.1f} obj/s]"
        self.__exit_event: Event = exit_event_queue
        self.__print_bar()

    def __del__(self) -> None:
        """
        Draw final bar
        """
        self.__print_bar(end="\n")

    def __print_bar(self, end: str = "") -> None:
        """
        Print progress bar to console
        """
        self.__write_time()
        self.__write_avg_speed()
        self.__write_percents_done()
        self.__write_preffix()
        p_bar = []
        p_bar.append(self.__progress)
        p_bar.append(self.__percents)
        p_bar.append(self.__avg_speed)
        p_bar.append(self.__time)
        p_bar_length = len("".join(p_bar))
        if p_bar_length < self.__terminal_columns:
            p_bar.insert(1, self.__fill_empty_space(p_bar_length))
        print(f"\r{''.join(p_bar)}", end=end)

    def __fill_empty_space(self, used_space: int) -> str:
        """
        Method to fill empty line space
        :param int used_space: already used space on terminal
        :rtype: str
        :return: string to fill current line to the end
        """
        available_space = self.__terminal_columns - used_space
        available_space -= 4
        space_done = available_space * self.data.current // self.data.total
        space_left = available_space - space_done
        results = "["
        results += self.data.markers.done * space_done
        in_progress = 0
        if space_left > 0:
            results += self.data.markers.current_marker()
            in_progress = 1
        results += self.data.markers.filler * (space_left - in_progress)
        results += "]"
        return results

    def __write_time(self) -> None:
        """
        Method to create running time string
        """
        r = self.data.times.running
        self.data.speed.update_speed_history()
        e = self.data.estimate
        self.__time = (
            f"[R {r[0]:02}:{r[1]:02}:{r[2]:02}][E {e[0]:02}:{e[1]:02}:{e[2]:02}]"
        )

    def __write_avg_speed(self) -> None:
        """
        Method to update avg speed
        """
        avg, u = convert_to_human(self.data.speed.avg_speed)
        if u:
            self.__avg_speed = f"[{avg:>5.1f}{u} obj/s]"
        else:
            self.__avg_speed = f"[{avg:>6.1f} obj/s]"

    def __write_percents_done(self) -> None:
        """
        Method to update percents done part
        """
        done = self.data.percents_done
        self.__percents = f"[{done:6.2f}%]"

    def __write_preffix(self) -> None:
        """
        Method to write preffix
        """
        down, d_u = convert_to_human(self.data.download)
        current, c_u = convert_to_human(self.data.current)
        if c_u:
            c_string = f"[V {current:>6.2f}{c_u}"
        else:
            c_string = f"[V {current:>7}"
        if d_u:
            d_string = f"|D {down:>6.2f}{d_u}|{self.__total}]"
        else:
            d_string = f"|D {down:>7}|{self.__total}]"
        self.__progress = c_string + d_string

    def draw_bar(self, update_interval: int = 1) -> None:
        """
        Method to draw bar in another thread
        """
        while True:
            try:
                self.__exit_event.wait(update_interval - 0.001)
            except EOFError:
                break
            self.__print_bar()
            if self.__exit_event.is_set():
                break



"""
Module for helping functions and classes
"""

import grp
import os
import pwd
from logging import getLogger
from typing import Tuple, Union

from s3ben.constants import SIZE_UNITS, UNITS

_logger = getLogger(__name__)


def drop_privileges(user: str) -> None:
    """
    Drop user privileges.

    :params str user: Username to which we should change permissions
    """
    new_user = pwd.getpwnam(user)
    if new_user.pw_uid == os.getuid():
        return
    new_gids = [new_user.pw_gid]
    new_gids += [
        group.gr_gid for group in grp.getgrall() if new_user.pw_name in group.gr_mem
    ]
    os.setgroups(new_gids[: os.NGROUPS_MAX])
    os.setgid(new_user.pw_uid)
    os.setuid(new_user.pw_uid)
    os.environ["HOME"] = new_user.pw_dir


def convert_to_human_v2(value: Union[int, float]) -> str:
    """
    Convert size to human
    """
    suffix = "B"
    for unit in SIZE_UNITS:
        if abs(value) < 1024.0:
            break
        if unit == UNITS[-1]:
            break
        value /= 1024.0
    return f"{value:3.2f}{unit}{suffix}"


def convert_to_human(value: Union[int, float]) -> tuple:
    """
    Onother convert function, should be renamed or consolidated
    """
    if float(value) <= 1000.0:
        return value, ""
    for unit in UNITS:
        value /= 1000.0
        if float(value) < 1000.0:
            break
        if unit == UNITS[-1]:
            break
    return value, unit


def check_object(path) -> Union[str, dict]:
    """
    Check object path for forward slash and replace
    if found as first character
    :param str path: Object path from s3
    """
    if path[0] == "/":
        remmpaed_obj = "_forward_slash_" + path
        _logger.warning(
            "Forward slash found for object: %s, remmaping to: %s", path, remmpaed_obj
        )
        return {path[1:]: remmpaed_obj}
    return path


def remmaping_message(action: str, remap: dict, bucket: str) -> Tuple[str, str, dict]:
    """
    Function to create remmaping message

    :param str action: Update or delete action
    :param dict remap: Remmaping data
    :param str bucket: Name of the bucket
    """
    remapping_update = {"action": action}
    remapping_update.update({"data": {"bucket": bucket, "remap": remap}})  # type: ignore
    local_path = next(iter(remap.values()))
    object_path = "/" + next(iter(remap.keys()))
    return object_path, local_path, remapping_update



"""
Module to handle object remapping
"""

import json
import multiprocessing
import os
from datetime import datetime
from logging import getLogger
from multiprocessing.synchronize import Event as EventClass
from queue import Empty

from typing_extensions import TypeAlias

_logger = getLogger(__name__)


class ResolveRemmaping:
    """
    Class to resolve remmapped objects
    :param str backup_root: Path to backup root
    """

    Queue: TypeAlias = multiprocessing.Queue
    Event: TypeAlias = EventClass

    def __init__(self, backup_root: str):
        self._queue = None
        self._remapping_db = os.path.join(backup_root, ".remappings")
        self._remappings_deleted = os.path.join(backup_root, ".deleted-remappings")

    def update_remapping(self, bucket: str, remap: dict) -> None:
        """
        Method to update remapping database
        :param str bucket: bucket name for which remap should be added
        :peram dict remap: dictionary containing remapping information
        :return: None
        """
        b_remaps = {}
        if not os.path.exists(self._remapping_db):
            _logger.warning("Remapping db doesn't exists, creating")
            update = {bucket: remap}
            with open(file=self._remapping_db, mode="w", encoding="utf-8") as f:
                json.dump(update, f)
                return
        with open(file=self._remapping_db, mode="r", encoding="utf-8") as f:
            remappings: dict = json.load(f)
        if bucket in remappings.keys():
            b_remaps = remappings.pop(bucket)
        b_remaps.update(remap)
        remappings.update({bucket: b_remaps})
        with open(file=self._remapping_db, mode="w", encoding="utf-8") as f:
            json.dump(obj=remappings, fp=f)

    def move_remapping(self, bucket: str, remap: dict) -> None:
        """
        Method to update remmaping database and remove deleted objects
        """
        if not os.path.exists(self._remapping_db):
            _logger.warning("Remapping database doesn't exists")
            return

        current_date = datetime.now().strftime("%Y-%m-%d")
        remapped_file = next(iter(remap.values()))
        remapped_key = next(iter(remap.keys()))
        _logger.debug("Moving %s to deleted items", remapped_file)

        with open(file=self._remapping_db, mode="r", encoding="utf-8") as f:
            remappings: dict = json.load(f)
        if remapped_key in remappings[bucket].keys():
            remappings[bucket].pop(remapped_key)
        else:
            _logger.warning("%s not found in remapping db", remapped_key)

        with open(file=self._remapping_db, mode="w", encoding="utf-8") as f:
            json.dump(obj=remappings, fp=f)
        if not os.path.exists(self._remappings_deleted):
            moved_dict = {
                current_date: {bucket: remap},
            }
            with open(file=self._remappings_deleted, mode="w", encoding="utf-8") as f:
                json.dump(moved_dict, f)
                return
        with open(file=self._remappings_deleted, mode="r", encoding="utf-8") as f:
            moved_remappings = json.load(f)
        moved_remappings[current_date][bucket].update(remap)
        with open(file=self._remappings_deleted, mode="w", encoding="utf-8") as f:
            json.dump(moved_remappings, f)

    def delete_remapping(self, key: str) -> None:
        """
        Method to remove moved remapping from deleted items

        :param str key: Key to remove from deleted remapping db
        """
        if not os.path.exists(self._remappings_deleted):
            _logger.warning("Deleted remapping doesn't exists")
            return
        with open(file=self._remappings_deleted, mode="r", encoding="utf-8") as f:
            remappings = json.load(f)
        if key in remappings.keys():
            _logger.info("removing %s key from deleted remappings", key)
            remappings.pop(key)
            with open(file=self._remappings_deleted, mode="w", encoding="utf-8") as f:
                json.dump(remappings, f)

    def run(self, queue: Queue, event: Event) -> None:
        """
        Method to launch Resolver as a process
        :param Queue queue: multiprocess.Queue class for receiving data
        :param Event event: multiprocess.Event class for receiving end event
        :return: None

        for updating remap db, dictionary must be added to queue:
        {
          "action": "update",
          "data": {
            "bucket": "bucket_name",
            "remap": {
              "object_key_to_match_local_file": "relative/path_to_key"
            }
          }
        }
        """
        _logger.info("starting remapping resolver")
        while not event.is_set():
            try:
                data: dict = queue.get(timeout=1)
            except Empty:
                if event.is_set():
                    _logger.debug("End event received")
                    break
                continue
            bucket = data["data"].get("bucket")
            remap = data["data"].get("remap")
            if data.get("action") in ["update", "download"]:
                self.update_remapping(bucket=bucket, remap=remap)
            if data.get("action") == "remove":
                remap = data.get("data")
                self.move_remapping(bucket=bucket, remap=remap)



"""
Backup manager module
"""

import hashlib
import multiprocessing
import os
import shutil
import threading
import time
from datetime import datetime, timedelta
from logging import getLogger
from multiprocessing.synchronize import Event as EventClass
from queue import Empty
from typing import Optional, Union

from tabulate import tabulate  # type: ignore
from typing_extensions import TypeAlias

from s3ben.helpers import (
    check_object,
    convert_to_human,
    convert_to_human_v2,
    drop_privileges,
)
from s3ben.progress import ConsoleBar
from s3ben.rabbit import MqConnection, RabbitMQ
from s3ben.remapper import ResolveRemmaping
from s3ben.s3 import S3Events
from s3ben.ui import S3benGui

_logger = getLogger(__name__)
Queue: TypeAlias = multiprocessing.Queue
Event: TypeAlias = EventClass


class BackupManager:
    """
    Class to coordinate all tasks

    :param str backup_root: Destination directory were all files will be placed
    :param str user: username to which change privileges
    :param str mq_queue: rabbitmq queue name
    :param RabbitMQ mq: RabbitMQ class object
    """

    def __init__(
        self,
        backup_root: str,
        user: str,
        mq_queue: Optional[str] = None,
        mq_conn: Optional[MqConnection] = None,
        s3_client: Optional[S3Events] = None,
        curses: bool = False,
    ):
        self._backup_root = backup_root
        self._user = user
        self._mq: Union[RabbitMQ, None] = None
        self._mq_conn = mq_conn
        self._mq_queue = mq_queue
        self._s3_client = s3_client
        self._bucket_name: Union[str, None] = None
        self._page_size: Union[int, None] = None
        self._progress_queue = None
        self._download_queue = None
        self._verify_queue = None
        self._remapped_queue = None
        self._end_event = None
        self._barrier = None
        self._curses = curses
        self._delay = 0
        self._mp_data_queue: Union[Queue, None] = None
        self._mp_end_event: Union[Event, None] = None

    def start_consumer(
        self,
        s3_client: S3Events,
        mp_data_queue: Queue,
        mp_end_event: Event,
    ) -> None:
        """
        Method to run rabbit mq consume
        """
        self._mp_data_queue = mp_data_queue
        self._mp_end_event = mp_end_event
        self._s3_client = s3_client
        _logger.debug("Dropping privileges to %s", self._user)
        drop_privileges(user=self._user)
        self.__run_consumer()

    def __run_consumer(self) -> None:
        """
        Method to run mq consumer
        """
        self._mq = RabbitMQ(conn_params=self._mq_conn)
        while True:
            try:
                self._mq.consume(
                    queue=self._mq_queue,
                    s3_client=self._s3_client,
                    mp_data_queue=self._mp_data_queue,
                    mp_end_event=self._mp_end_event,
                )
            except (KeyboardInterrupt, SystemExit):
                _logger.debug("Stopping mq")
                self._mq.stop()
                break
            self.__reconnect_consumer()

    def __reconnect_consumer(self) -> None:
        """
        Method to reconnect consumer
        """
        if self._mq.should_reconnect:
            self._mq.stop()
            r_delay = self.__reconnect_delay()
            _logger.warning("Connection lost, reconnecting after %s seconds", r_delay)
            time.sleep(r_delay)
            self._mq = RabbitMQ(conn_params=self._mq_conn)

    def __reconnect_delay(self) -> int:
        """
        Method to manage reconnect delay,
        current max delay 5s
        :rtype: int
        :return: Number of seconds to sleep
        """
        max_delay = 5
        if self._mq.was_consuming:
            self._delay = 0
        else:
            self._delay += 1
        return min(max_delay, self._delay)

    def __progress2(self, avg_interval: int, update_interval: int) -> None:
        """
        Updated progress bar
        """
        info = self._s3_client.get_bucket(self._bucket_name)
        total_objects = info["usage"]["rgw.main"]["num_objects"]
        progress = ConsoleBar(
            total=total_objects,
            avg_interval=avg_interval,
            exit_event_queue=self._end_event,
        )
        decoration = threading.Thread(target=progress.draw_bar, args=(update_interval,))
        decoration.start()
        while True:
            try:
                data = self._progress_queue.get(block=True, timeout=0.1)
            except Empty:
                if self._end_event.is_set():
                    break
                continue
            else:
                progress.data.update_counters(data)

    def sync_bucket(
        self,
        bucket_name: str,
        threads: int,
        page_size: int,
        checkers: int,
        skip_checksum: bool,
        skip_filesize: bool,
        avg_interval: int,
        update_interval: int,
    ) -> None:
        _logger.info("Starting bucket sync")
        start = time.perf_counter()
        self._page_size = page_size
        self._bucket_name = bucket_name
        proc_manager = multiprocessing.managers.SyncManager()
        proc_manager.start()
        self._download_queue = proc_manager.Queue(maxsize=threads * 2)
        self._verify_queue = proc_manager.Queue(maxsize=threads * 2)
        self._progress_queue = proc_manager.Queue()
        self._remapped_queue = proc_manager.Queue()
        self._end_event = proc_manager.Event()
        self._barrier = proc_manager.Barrier(threads + checkers)
        try:
            reader = multiprocessing.Process(target=self._page_reader)
            reader.start()
            remapper = multiprocessing.Process(target=self.__remapper)
            remapper.start()
            processess = []
            for _ in range(checkers):
                verify = multiprocessing.Process(
                    target=self._page_verfication,
                    args=(
                        skip_checksum,
                        skip_filesize,
                    ),
                )
                processess.append(verify)
            for _ in range(threads):
                download = multiprocessing.Process(target=self._page_processor)
                processess.append(download)
            for proc in processess:
                proc.start()
            processess.append(reader)
            processess.append(remapper)
            if not self._curses:
                self.__progress2(
                    avg_interval=avg_interval, update_interval=update_interval
                )
            else:
                self._curses_ui()
            for proc in processess:
                proc.join()
        except KeyboardInterrupt:
            for proc in processess:
                proc.terminate()
        finally:
            proc_manager.shutdown()
        end = time.perf_counter()
        _logger.info("Sync took: %s seconds", round(end - start, 2))

    def _curses_ui(self) -> None:
        info = self._s3_client.get_bucket(self._bucket_name)
        total_objects = info["usage"]["rgw.main"]["num_objects"]
        ui = S3benGui(title=f"Syncing buclet: {self._bucket_name}", total=total_objects)
        while True:
            try:
                data = self._progress_queue.get(timeout=0.5)
            except Empty:
                ui.progress(progress=ui._progress)
                continue
            else:
                ui.progress(progress=ui._progress + data)

    def __remapper(self) -> None:
        """
        Method to run remmaper class
        :return: None
        """
        remap_resolver = ResolveRemmaping(backup_root=self._backup_root)
        remap_resolver.run(queue=self._remapped_queue, event=self._end_event)

    def _page_reader(self) -> None:
        _logger.info("Starting page processing")
        self._end_event.clear()
        paginator = self._s3_client.client_s3.get_paginator("list_objects_v2")
        page_config = {"PageSize": self._page_size}
        pages = paginator.paginate(
            Bucket=self._bucket_name, PaginationConfig=page_config
        )
        for page in pages:
            self._verify_queue.put(page["Contents"])
        _logger.debug("Finished reading pages")
        self._end_event.set()

    def _page_verfication(self, skip_checksum: bool, skip_filesize: bool) -> None:
        """
        Method to verify file by comparing size and checksum
        :returns: None
        """
        _logger.debug("Running page verification")
        self._barrier.wait()
        while True:
            try:
                data = self._verify_queue.get(block=False)
            except Empty:
                if self._end_event.is_set():
                    _logger.debug("Braking from queue")
                    break
                continue
            download_list = []
            for obj in data:
                obj_key = obj.get("Key")
                key = check_object(obj_key)
                if isinstance(key, dict):
                    remapping_update = {"action": "update"}
                    local_path = next(iter(key.values()))
                    remapping_update.update(
                        {"data": {"bucket": self._bucket_name, "remap": key}}
                    )
                    self._remapped_queue.put(remapping_update)
                else:
                    local_path = key
                fp_key = os.path.join(
                    self._backup_root, "active", self._bucket_name, local_path
                )
                ob_key_exists = self.__check_file(path=fp_key)
                if not ob_key_exists:
                    _logger.debug("file doesn't exists: %s", fp_key)
                    download_list.append(key)
                    continue
                if not skip_checksum:
                    obj_sum = obj.get("ETag")
                    obj_sum_matches = self.__check_md5(path=fp_key, md5=obj_sum)
                    if not obj_sum_matches:
                        download_list.append(key)
                        continue
                if not skip_filesize:
                    obj_size = obj.get("Size")
                    obj_size_matches = self.__check_file_size(
                        path=fp_key, size=obj_size
                    )
                    if not obj_size_matches:
                        download_list.append(key)
                        continue
            skipped = len(data) - len(download_list)
            progress_update = {"vrf": skipped}
            self._progress_queue.put(progress_update)
            if len(download_list) > 0:
                self._download_queue.put(download_list)

    def __check_file_size(self, path: str, size: int) -> bool:
        """
        Method to check if file size matches
        :param str path: full path to the file
        :param int size: size of remote object to verify
        :return: True if matches, otherwise False
        """
        _logger.debug("Checking file size %s", path)
        local_size = os.stat(path=path).st_size
        if local_size == size:
            return True
        return False

    def __check_file(self, path: str) -> bool:
        """
        Method to check if local file exists
        :return: True if file exisrts, otherwise false
        """
        _logger.debug("Checking if file exists: %s", path)
        if os.path.isfile(path=path):
            return True
        return False

    def __check_md5(self, path: str, md5: str) -> bool:
        """
        Method to calculate file md5 sum and check if it matches
        :param str path: full path to the local file
        :param str md5: md5 to check against
        :return: True if sum maches, otherwise false
        """
        _logger.debug("Checking md5sum for %s", path)
        with open(path, "rb") as file:
            source_md5 = hashlib.md5()
            while chunk := file.read(8192):
                source_md5.update(chunk)
        calculated_md5 = source_md5.hexdigest()
        if calculated_md5 == md5.replace('"', ""):
            return True
        return False

    def _page_processor(self) -> None:
        proc = multiprocessing.current_process().name
        _logger.debug("Running: %s", proc)
        self._barrier.wait()
        while True:
            try:
                data = self._download_queue.get(block=True, timeout=0.2)
            except Empty:
                if self._end_event.is_set():
                    break
                continue
            else:
                self._s3_client.download_all_objects(self._bucket_name, data)
                progress_update = {"dl": len(data)}
                self._progress_queue.put(progress_update)

    def list_buckets(
        self,
        exclude: list,
        show_excludes: bool,
        show_obsolete: bool,
        only_enabled: bool,
        sort: str,
        sort_revers: bool,
    ) -> None:
        results = []
        s3_buckets = self._s3_client.get_admin_buckets()
        s3ben_buckets = os.listdir(os.path.join(self._backup_root, "active"))
        merged_list = list(dict.fromkeys(s3_buckets + s3ben_buckets))
        for bucket in merged_list:
            bucket_excluded = True if bucket in exclude else ""
            enabled = True if bucket in s3ben_buckets else ""
            obsolete = True if bucket not in s3_buckets else ""
            if not show_excludes and bucket_excluded:
                continue
            if not show_obsolete and obsolete:
                continue
            if only_enabled and not enabled:
                continue
            remote_size = 0
            objects = 0
            unit = ""
            if bucket in s3_buckets:
                bucket_info = self._s3_client.get_bucket(bucket=bucket)
                if "rgw.main" in bucket_info["usage"].keys():
                    original_size = bucket_info["usage"]["rgw.main"].get(
                        "size_utilized"
                    )
                    remote_size = convert_to_human_v2(original_size)
                    original_objects = bucket_info["usage"]["rgw.main"].get(
                        "num_objects"
                    )
                    objects, unit = convert_to_human(original_objects)
            remote_format = ">3d" if isinstance(objects, int) else ">5.2f"
            info = {
                "Bucket": bucket,
                "Owner": bucket_info.get("owner"),
            }
            if not only_enabled:
                info["Enabled"] = enabled
            info.update(
                {
                    "size": original_size,
                    "Remote size": remote_size,
                    "objects": original_objects,
                    "Remote objects": f"{objects:{remote_format}}{unit}",
                }
            )
            if show_excludes and not only_enabled:
                info["Exclude"] = bucket_excluded

            if show_obsolete and not only_enabled:
                info["Obsolete"] = obsolete
            results.append(info)
        if sort == "bucket":
            results = sorted(results, key=lambda k: k["Bucket"], reverse=sort_revers)
        if sort == "owner":
            results = sorted(results, key=lambda k: k["Owner"], reverse=sort_revers)
        if sort == "size":
            results = sorted(results, key=lambda k: k["size"], reverse=sort_revers)
        if sort == "objects":
            results = sorted(results, key=lambda k: k["objects"], reverse=sort_revers)
        for r in results:
            r.pop("size")
            r.pop("objects")
        print(tabulate(results, headers="keys"))

    def cleanup_deleted_items(self, days: int) -> None:
        """
        Method to cleanup deleted items
        :param int days: Days to keep deleted items
        :return: None
        """
        cleanup_remap = ResolveRemmaping(backup_root=self._backup_root)
        deleted_path = os.path.join(self._backup_root, "deleted")
        date_to_remove = datetime.now() - timedelta(days=days)
        _, dirs, _ = next(os.walk(deleted_path))
        for d in dirs:
            dir_date = datetime.strptime(d, "%Y-%m-%d")
            if dir_date > date_to_remove:
                continue
            cleanup_remap.delete_remapping(key=d)
            to_remove = os.path.join(deleted_path, d)
            _logger.info("Removing %s", to_remove)
            shutil.rmtree(path=to_remove)
        _logger.debug("Cleanup done")
